using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.IO;

namespace FolderCp
{
	internal partial class FilterForm : Form
	{
		internal FilterItem fi;

		internal FilterForm(FilterItem src)
		{
			InitializeComponent();
			fi = new FilterItem();

			if (src != null)
			{
				fi.action = src.action;
				fi.type = src.type;
				fi.pattern = src.pattern;

				switch(fi.action)
				{
//					case FilterAction.EXCLUDE:
//						rbExclude.Checked = true;
//						break;
					case FilterAction.INCLUDE:
						rbInclude.Checked = true;
						break;
					case FilterAction.DISABLED:
						rbDisabled.Checked = true;
						break;
				}

				if (fi.type == FilterType.WILDCARD)
					rbWildcard.Checked = true;

				tbPattern.Text = src.pattern;
			}			
		}

		private static void MsgBox(string msg)
		{
			MessageBox.Show(msg, "Error");
		}

		private bool MakeFI()
		{
			fi.pattern = tbPattern.Text.Trim();
			if (String.IsNullOrEmpty(fi.pattern))
			{
				MsgBox("Empty pattern");
				return false;
			}

			if (rbExclude.Checked)
				fi.action = FilterAction.EXCLUDE;
			else if (rbInclude.Checked)
				fi.action = FilterAction.INCLUDE;
			else if (rbDisabled.Checked)
				fi.action = FilterAction.DISABLED;
			else
				return false;

			if (rbRegex.Checked)
				fi.type = FilterType.REGEX;
			else if (rbWildcard.Checked)
				fi.type = FilterType.WILDCARD;
			else
				return false;

			return true;
		}

		private void btnOK_Click(object sender, EventArgs e)
		{
			if (MakeFI())
			{
				DialogResult = DialogResult.OK;
				Close();
			}
		}

		private void btnCancel_Click(object sender, EventArgs e)
		{
			DialogResult = DialogResult.Cancel;
			Close();
		}

		private void btnMatch_Click(object sender, EventArgs e)
		{
			if(MakeFI())
			{
				string test = tbTest.Text.Trim();
				if(String.IsNullOrEmpty(test))
				{
					MsgBox("Empty test string.");
					return;
				}

				if(test.IndexOf("*") >= 0 || test.IndexOf('?') >= 0 || test.IndexOf(@"\") >= 0)
				{
					MsgBox("Invalid filename letter in test string.");
					return;
				}

				CompiledFilter cf = new CompiledFilter(fi);
				MessageBox.Show(cf.Match(test) ? "Matched" : "Not matched", "Result");
			}
		}
	};

	internal enum FilterAction
	{
		EXCLUDE = 1,
		INCLUDE = 2,
		DISABLED = 3
	}

	internal enum FilterType
	{
		REGEX = 1,
		WILDCARD = 2
	};

	internal class FilterItem
	{
		internal FilterAction action;
		internal FilterType type;
		internal string pattern;
		internal ListViewItem _lvi;

		internal string GetActionName()
		{
			switch (action)
			{
				case FilterAction.EXCLUDE:
					return "Exclude";
				case FilterAction.INCLUDE:
					return "Include";
				case FilterAction.DISABLED:
					return "Disabled";
				default:
					return "(ERROR)";
			}
		}

		internal string GetTypeName()
		{
			switch (type)
			{
				case FilterType.REGEX:
					return "RegEx";
				case FilterType.WILDCARD:
					return "Wildcard";
				default:
					return "(ERROR)";
			}
		}

		internal static FilterItem Parse(string spec)
		{
			FilterItem item = new FilterItem();

			switch (spec[0])
			{
				case 'X':
					item.action = FilterAction.EXCLUDE;
					break;
				case 'I':
					item.action = FilterAction.INCLUDE;
					break;
				case 'D':
					item.action = FilterAction.DISABLED;
					break;
				default:
					return null;
			}

			switch (spec[1])
			{
				case 'R':
					item.type = FilterType.REGEX;
					break;
				case 'W':
					item.type = FilterType.WILDCARD;
					break;
				default:
					return null;
			}

			if (spec[2] != ':')
				return null;

			item.pattern = spec.Substring(3);

			return item;
		}

		internal string Unparse()
		{
			char _type, _action;

			switch (action)
			{
				case FilterAction.DISABLED:
					_action = 'D';
					break;

				case FilterAction.EXCLUDE:
					_action = 'X';
					break;

				case FilterAction.INCLUDE:
					_action = 'I';
					break;

				default:
					return null;
			}

			switch (type)
			{
				case FilterType.REGEX:
					_type = 'R';
					break;

				case FilterType.WILDCARD:
					_type = 'W';
					break;

				default:
					return null;
			}

			return String.Format("{0}{1}:{2}", _action, _type, pattern);
		}
	};

	internal class CompiledFilter
	{
		private Regex m_re;
		private bool m_IsExclude;

		internal CompiledFilter(FilterItem src)
		{
			m_IsExclude = (src.action == FilterAction.EXCLUDE);

			if(src.type == FilterType.REGEX)
			{
				m_re = new Regex(src.pattern.ToLower());
			}
			else
			{
				// wildcard matching
				string pattern = src.pattern.ToLower();

				pattern = pattern.Replace(".", @"\.");
				pattern = pattern.Replace('?', '.').Replace("*", ".*");

				m_re = new Regex("^" + pattern + "$");
			}
		}

		internal bool IsExclude
		{
			get { return m_IsExclude; }
		}

		internal bool Match(string eval)
		{
			eval = eval.Replace('\\', '/').ToLower();
			if (eval.IndexOf('/') >= 0)
				return m_re.IsMatch(Path.GetFileName(eval));
			else
				return m_re.IsMatch(eval);
		}
	}
}