using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Collections;
using System.Diagnostics;
using System.Security.Cryptography;

namespace FolderCp
{
	enum ReportCategory
	{
		COPY,
		SKIP,
		EXCLUDE
	}

	internal delegate void DelegateLogMsg(string msg);
	internal delegate void DelegatePurgeFile(string filename, long filesize);
	internal delegate void DelegateReportFile(string filename, long filesize, ReportCategory cate);

	internal partial class MainForm : Form
	{
		private const string APPTITLE = "Smart Folder Copy";
		internal static MainForm s_this;

		private bool ExitOnBgFinish;
		private bool IsSimulation;
		private string CurrentWorkDir;

		internal DelegateLogMsg dele_LogMsg;
		internal DelegatePurgeFile dele_PurgeFile;
		internal DelegateReportFile dele_ReportFile;

		internal MainForm()
		{
			s_this = this;

			InitializeComponent();
			LogAction("Program Ready.");
			PresetFilename = null;

#if DEBUG
			LoadPreset(@"d:\work\foldercp\test.fcp");
#endif

			dele_LogMsg = new DelegateLogMsg(_LogMsg);
			dele_PurgeFile = new DelegatePurgeFile(_PurgeFile);
			dele_ReportFile = new DelegateReportFile(_ReportFile);

			AdjustColumnWidth(lvCopy);
			AdjustColumnWidth(lvSkip);
			AdjustColumnWidth(lvExclude);
			AdjustColumnWidth(lvDelete);
		}

		private void AdjustColumnWidth(ListView lv)
		{
			lv.Columns[0].Width = lv.Width - 96;
			lv.Columns[1].Width = 96;
		}

		private string m_PresetFilename;
		private string PresetFilename
		{
			get { return m_PresetFilename; }
			set 
			{
				m_PresetFilename = value;
				this.Text = APPTITLE;
				if (value != null)
					this.Text += " - " + m_PresetFilename;
			}
		}

		private void _LogMsg(string msg)
		{
			LogAction(msg);
		}

		private void _PurgeFile(string filename, long filesize)
		{
			ListViewItem lvi = lvDelete.Items.Add(filename);
			lvi.SubItems.Add("Not Deleted");
		}

		private void _ReportFile(string filename, long filesize, ReportCategory cate)
		{
			ListViewItem lvi;

			switch(cate)
			{
				case ReportCategory.COPY:
					lvi = lvCopy.Items.Add(filename);
					break;

				case ReportCategory.SKIP:
					lvi = lvSkip.Items.Add(filename);
					break;

				case ReportCategory.EXCLUDE:
					lvi = lvExclude.Items.Add(filename);
					break;

				default:
					MsgBox("Invalid file log cateogry:" + cate);
					return;
			}

			if(filesize < 0)
				lvi.SubItems.Add("<DIR>");
			else
				lvi.SubItems.Add(filesize.ToString());
		}

		private static void MsgBox(string msg)
		{
			MessageBox.Show(msg, APPTITLE);
		}

		private static DialogResult MsgBox(string msg, MessageBoxButtons mbb)
		{
			return MessageBox.Show(msg, APPTITLE, mbb);
		}

		private void mnuAbout_Click(object sender, EventArgs e)
		{
			MsgBox("Smart Folder Copy Version 0.1.2\r\n\r\nVisit http://foldercp.kldp.net");
		}

		private void mnuExit_Click(object sender, EventArgs e)
		{
			Close();
		}

		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if(BgWorker.IsBusy)
			{
				if (ExitOnBgFinish)
				{
					MsgBox("Now terminating, please wait.");
				}
				else
				{
					if (MsgBox("Copy process is not finished. Exit anyway?", MessageBoxButtons.YesNo) == DialogResult.Yes)
					{
						ExitOnBgFinish = true;
						BgWorker.CancelAsync();
					}
				}

				e.Cancel = true;
			}
		}

		private void LogAction(string msg)
		{
			ListViewItem itm = lvAction.Items.Add(String.Format("{0:MM-dd HH:mm.ss}", DateTime.Now));
			itm.SubItems.Add(msg);
		}

		private void btnBrowseSrc_Click(object sender, EventArgs e)
		{
			BrowseFolder(tbSrcFolder);
		}

		private void btnBrowseDst_Click(object sender, EventArgs e)
		{
			BrowseFolder(tbDstFolder);
		}

		private void BrowseFolder(TextBox tb)
		{
			FolderBrowserDialog dlg = new FolderBrowserDialog();

			if (tb == tbSrcFolder)
			{
				dlg.ShowNewFolderButton = false;
				dlg.Description = "Select the source folder to backup.";
			}
			else
			{
				dlg.ShowNewFolderButton = true;
				dlg.Description = "Select the destination folder to be written.";
			}

			dlg.RootFolder = Environment.SpecialFolder.MyComputer;
			dlg.SelectedPath = tb.Text;

			if (dlg.ShowDialog() == DialogResult.OK)
			{
				tb.Text = dlg.SelectedPath;
			}
		}

		private void AddFilter(FilterItem fi)
		{
			ListViewItem lvi = lvFilter.Items.Add(fi.GetActionName());
			lvi.SubItems.Add(fi.GetTypeName());
			lvi.SubItems.Add(fi.pattern);
			lvi.Tag = fi;
			fi._lvi = lvi;
		}

		private void UpdateFilter(FilterItem fi)
		{
			fi._lvi.Text = fi.GetActionName();
			fi._lvi.SubItems[1].Text = fi.GetTypeName();
			fi._lvi.SubItems[2].Text = fi.pattern;
		}

		private bool LoadPreset(string pathname)
		{
			string SRC = null;
			string DST = null;
			List<FilterItem> FILTERS = new List<FilterItem>();

			LogAction("Load preset: " + pathname);

			try
			{
				TextReader tr = new StreamReader(File.OpenRead(pathname));

				while (true)
				{
					string line = tr.ReadLine();
					if (line == null)
						break;

					line = line.Trim();
					if (String.IsNullOrEmpty(line))
						continue;

					int iEqual = line.IndexOf('=');
					string cmd = line.Substring(0, iEqual);
					string param = line.Substring(iEqual + 1);

					switch(cmd)
					{
						case "SRC":
							SRC = param;
							break;
						case "DST":
							DST = param;
							break;
						case "FILTER":
							FilterItem item = FilterItem.Parse(param);
							if(item == null)
							{
								LogAction("Filter parsing failed: " + param);
								return false;
							}
							FILTERS.Add(item);
							break;
						default:
							LogAction("Invalid command: " + cmd);
							return false;
					}
				}

				tr.Close();			
			}
			catch(Exception ex)
			{
				LogAction("Exception: " + ex.Message);
				return false;
			}

			tbSrcFolder.Text = SRC;
			tbDstFolder.Text = DST;
			
			lvFilter.Items.Clear();
			foreach(FilterItem fi in FILTERS)
			{
				AddFilter(fi);
			}

			PresetFilename = pathname;
			return true;
		}

		private void SavePreset(string pathname)
		{
			LogAction("Save preset: " + pathname);

			StringBuilder sb = new StringBuilder();
			sb.AppendFormat("SRC={0}\n", tbSrcFolder.Text);
			sb.AppendFormat("DST={0}\n", tbDstFolder.Text);

			foreach(ListViewItem lvi in lvFilter.Items)
			{
				FilterItem fi = (FilterItem) lvi.Tag;

				string spec = fi.Unparse();
				if(spec == null)
				{
					MsgBox("Filter error.");
					return;
				}

				sb.AppendFormat("FILTER={0}\n", spec);
			}

			TextWriter tw = new StreamWriter(File.OpenWrite(pathname));
			tw.Write(sb);
			tw.Close();

			PresetFilename = pathname;
		}

		private void mnuLoadPreset_Click(object sender, EventArgs e)
		{
			OpenFileDialog dlg = new OpenFileDialog();

			dlg.ShowReadOnly = true;
			dlg.Multiselect = false;
			dlg.CheckFileExists = true;
			dlg.FileName = PresetFilename;

			if (dlg.ShowDialog() == DialogResult.OK)
			{
				LoadPreset(dlg.FileName);
			}
		}

		private void mnuSavePreset_Click(object sender, EventArgs e)
		{
			if(!String.IsNullOrEmpty(PresetFilename))
			{
				DialogResult r = MsgBox("Save preset to '" + PresetFilename + "'?", MessageBoxButtons.YesNoCancel);
				if (r == DialogResult.Cancel)
					return;

				if (r == DialogResult.Yes)
				{
					SavePreset(PresetFilename);
					return;
				}
			}

			SaveFileDialog dlg = new SaveFileDialog();
			dlg.FileName = PresetFilename;
			dlg.CheckPathExists = true;

			if(dlg.ShowDialog() == DialogResult.OK)
			{
				SavePreset(dlg.FileName);
			}

		}

		private void lvFilter_MouseUp(object sender, MouseEventArgs e)
		{
			if(BgWorker.IsBusy)
				return;

			if (e.Button == MouseButtons.Right)
			{
				mnuRemoveFilter.Enabled =
				mnuDisableFilter.Enabled =
				mnuEditFilter.Enabled = lvFilter.SelectedItems.Count > 0;

				FilterCtxMenu.Show(MousePosition);
			}
		}

		private void mnuAddFilter_Click(object sender, EventArgs e)
		{
			FilterForm form = new FilterForm(null);
			if(form.ShowDialog() == DialogResult.OK)
			{
				AddFilter(form.fi);
			}
		}

		private FilterItem GetSelectedFilterItem()
		{
			Debug.Assert(lvFilter.SelectedItems.Count == 1);
			foreach(ListViewItem lvi in lvFilter.SelectedItems)
				return (FilterItem) lvi.Tag;

			return null;
		}

		private void mnuRemoveFilter_Click(object sender, EventArgs e)
		{
			FilterItem fi = GetSelectedFilterItem();
			if (fi == null) return;

			lvFilter.Items.Remove(fi._lvi);
		}

		private void mnuDisableFilter_Click(object sender, EventArgs e)
		{
			FilterItem fi = GetSelectedFilterItem();
			if (fi == null) return;

			fi.action = FilterAction.DISABLED;
			UpdateFilter(fi);
		}

		private void mnuEditFilter_Click(object sender, EventArgs e)
		{
			FilterItem fi = GetSelectedFilterItem();
			if (fi == null) return;

			FilterForm form = new FilterForm(fi);
			if(form.ShowDialog() == DialogResult.OK)
			{
				fi.action = form.fi.action;
				fi.type = form.fi.type;
				fi.pattern = form.fi.pattern;
				UpdateFilter(fi);
			}
		}

		private void mnuDelete_Click(object sender, EventArgs e)
		{
			MsgBox("NOT IMPL");
		}

		#region Copy Worker Thread

		internal static List<CompiledFilter> CompiledFilters;

		private bool PrepareWork()
		{
			if (!Directory.Exists(tbSrcFolder.Text))
			{
				MsgBox("Source folder does not exist.");
				return false;
			}

			string DstFolder = tbDstFolder.Text = tbDstFolder.Text.Trim();
			if (String.IsNullOrEmpty(DstFolder))
			{
				MsgBox("Input the destination path.");
				return false;
			}

			if (!Directory.Exists(DstFolder))
			{
				if (MsgBox("Destination folder doesn't exist. Create?", MessageBoxButtons.YesNo) != DialogResult.Yes)
				{
					MsgBox("Copy process aborted.");
					return false;
				}

				try
				{
					Directory.CreateDirectory(DstFolder);
				}
				catch (Exception)
				{
					MsgBox("CreateDirectory failed.");
					return false;
				}
			}

			if (CompiledFilters == null)
				CompiledFilters = new List<CompiledFilter>();
			else
				CompiledFilters.Clear();

			foreach (ListViewItem lvi in lvFilter.Items)
			{
				FilterItem fi = (FilterItem)lvi.Tag;
				if (fi.action == FilterAction.DISABLED)
					continue;

				CompiledFilters.Add(new CompiledFilter(fi));
			}

			mnuLoadPreset.Enabled =
			mnuSimulate.Enabled =
			mnuCopy.Enabled =
			btnBrowseDst.Enabled =
			btnBrowseSrc.Enabled =
			tbDstFolder.Enabled =
			tbSrcFolder.Enabled = false;
			mnuCancel.Enabled = true;

			ExitOnBgFinish = false;

			lvCopy.Items.Clear();
			lvDelete.Items.Clear();
			lvExclude.Items.Clear();
			lvSkip.Items.Clear();

			StatusLabel.Text = "Scanning source folder...";
			return true;
		}

		private void mnuSimulate_Click(object sender, EventArgs e)
		{
			Debug.Assert(BgWorker.IsBusy == false);
			if (PrepareWork())
			{
				IsSimulation = true;
				BgWorker.RunWorkerAsync();
			}
		}

		private void mnuCopy_Click(object sender, EventArgs e)
		{
			Debug.Assert(BgWorker.IsBusy == false);
			if (PrepareWork())
			{
				IsSimulation = false;
//				BgWorker_DoWork(sender,null);	// easy debug
				BgWorker.RunWorkerAsync();
			}
		}

		private void mnuCancel_Click(object sender, EventArgs e)
		{
			if (MsgBox("Cancel the copy process?", MessageBoxButtons.YesNo) == DialogResult.Yes)
			{
				if (BgWorker.IsBusy)
					BgWorker.CancelAsync();
			}
		}

		private void BgWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
		{
			StatusLabel.Text = CurrentWorkDir;
			StatusProgress.Value = e.ProgressPercentage;
		}

		private void BgWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
		{
			if (ExitOnBgFinish)
			{
				Application.Exit();
				return;
			}

			StatusLabel.Text = "Done.";

			mnuLoadPreset.Enabled =
			mnuSimulate.Enabled =
			mnuCopy.Enabled =
			btnBrowseDst.Enabled =
			btnBrowseSrc.Enabled =
			tbDstFolder.Enabled =
			tbSrcFolder.Enabled = true;
			mnuCancel.Enabled = false;
		}

		private int numSrcFiles;
		private int numProcessed;

		private void CopyFile(FileInfo src, FileInfo dst)
		{
			Invoke(dele_ReportFile, new Object[] { src.FullName, src.Length, ReportCategory.COPY });

			if (!IsSimulation)
			{
				// TODO: real file copy
			}
		}

		private void CopyFolder(Folder src, string absdest)
		{
			// no dest exists, copy all from src.

			foreach (string fname in src.Files)
			{
				if (BgWorker.CancellationPending)
					return;

				numProcessed++;
				BgWorker.ReportProgress(numProcessed * 100 / numSrcFiles);

				FileInfo sf = new FileInfo(Path.Combine(src.AbsPath, fname));
				FileInfo df = new FileInfo(Path.Combine(absdest, fname));
				CopyFile(sf, df);
			}

			foreach (Folder sub in src.Dirs.Values)
			{
				if (BgWorker.CancellationPending)
					return;

				CopyFolder(sub, Path.Combine(absdest, sub.Name));
			}
		}

		private void ProcessFolder(Folder src, string absdest)
		{
			DiffResult diff = src.MakeDiff(absdest);

			CurrentWorkDir = src.RelPath;

			// check if two files are same, copy if different.
			foreach (string fname in diff.CommonFiles)
			{
				if (BgWorker.CancellationPending)
					return;

				numProcessed++;
				BgWorker.ReportProgress(numProcessed * 100 / numSrcFiles);

				FileInfo sf = new FileInfo(Path.Combine(src.AbsPath, fname));
				FileInfo df = new FileInfo(Path.Combine(absdest, fname));

				if (sf.Length == df.Length)
				{
					// md5 check
					byte[] shash = GetChecksum(sf);
					byte[] dhash = GetChecksum(df);

					if (shash.Equals(dhash))
					{
						// no need to copy
						Invoke(dele_ReportFile, new Object[] { sf.FullName, sf.Length, ReportCategory.SKIP });
						continue;
					}
				}

				CopyFile(sf, df);
			}

			foreach (string fname in diff.SrcOnlyFiles)
			{
				if (BgWorker.CancellationPending)
					return;

				numProcessed++;
				BgWorker.ReportProgress(numProcessed * 100 / numSrcFiles);

				FileInfo sf = new FileInfo(Path.Combine(src.AbsPath, fname));
				FileInfo df = new FileInfo(Path.Combine(absdest, fname));
				CopyFile(sf, df);
			}

			foreach (string fname in diff.DstOnlyFiles)
			{
				if (BgWorker.CancellationPending)
					return;

//				LogAction("Delete: " + Path.Combine(absdest, fname));
			}

			foreach (string dname in diff.CommonDirs)
			{
				if (BgWorker.CancellationPending)
					return;

				ProcessFolder(src.Dirs[dname], Path.Combine(absdest, dname));
			}

			foreach (string dname in diff.SrcOnlyDirs)
			{
				if (BgWorker.CancellationPending)
					return;

				CopyFolder(src.Dirs[dname], Path.Combine(absdest, dname));
			}
		}

		private static byte[] GetChecksum(FileInfo fi)
		{
			FileStream fs = fi.OpenRead();
			MD5CryptoServiceProvider csp = new MD5CryptoServiceProvider();
			//SHA1CryptoServiceProvider csp = new SHA1CryptoServiceProvider();

			byte[] hash = csp.ComputeHash(fs);
			fs.Close();

			return hash;
		}

		private void BgWorker_DoWork(object sender, DoWorkEventArgs e)
		{
			string SrcBase = tbSrcFolder.Text;
			string DstBase = tbDstFolder.Text;

			try
			{
				DirectoryInfo diSrc = new DirectoryInfo(SrcBase);
				DirectoryInfo diDst = new DirectoryInfo(DstBase);
				CurrentWorkDir = "/";

				// scan source folder recursively
				Folder srcRoot = new Folder(null, null, SrcBase);
				numSrcFiles = srcRoot.GetFileCount();
				numProcessed = 0;

				ProcessFolder(srcRoot, DstBase);
			}
			catch (Exception ex)
			{
				MsgBox(ex.ToString());
			}
		}

		#endregion	// Copy Worker Thread
	}
}